// GMEDoc.cpp : implementation of the CGMEDoc class
//

#include "stdafx.h"

#include "GMEApp.h"
#include "GMEstd.h"

#include "Mainfrm.h"
#include "GMEDoc.h"
#include "GMEView.h"
#include "GuiObject.h"
#include "GMEEventLogger.h"
#include "Parser.h"

#include "slist"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CGMEDoc

CGMEDoc* CGMEDoc::theInstance = 0;

IMPLEMENT_DYNCREATE(CGMEDoc, CDocument)

BEGIN_MESSAGE_MAP(CGMEDoc, CDocument)
	//{{AFX_MSG_MAP(CGMEDoc)
	ON_COMMAND(ID_MODE_AUTOCONNECT, OnModeAutoconnect)
	ON_UPDATE_COMMAND_UI(ID_MODE_AUTOCONNECT, OnUpdateModeAutoconnect)
	ON_COMMAND(ID_MODE_DISCONNECT, OnModeDisconnect)
	ON_UPDATE_COMMAND_UI(ID_MODE_DISCONNECT, OnUpdateModeDisconnect)
	ON_COMMAND(ID_MODE_EDIT, OnModeEdit)
	ON_UPDATE_COMMAND_UI(ID_MODE_EDIT, OnUpdateModeEdit)
	ON_COMMAND(ID_MODE_SET, OnModeSet)
	ON_UPDATE_COMMAND_UI(ID_MODE_SET, OnUpdateModeSet)
	ON_COMMAND(ID_MODE_VISUALIZE, OnModeVisualize)
	ON_UPDATE_COMMAND_UI(ID_MODE_VISUALIZE, OnUpdateModeVisualize)
	ON_COMMAND(ID_MODE_ZOOM, OnModeZoom)
	ON_UPDATE_COMMAND_UI(ID_MODE_ZOOM, OnUpdateModeZoom)
	ON_COMMAND(ID_MODE_SHORTCONNECT, OnModeShortConnect)
	ON_UPDATE_COMMAND_UI(ID_MODE_SHORTCONNECT, OnUpdateModeShortConnect)
	ON_COMMAND(ID_MODE_SHORTDISCONNECT, OnModeShortDisconnect)
	ON_UPDATE_COMMAND_UI(ID_MODE_SHORTDISCONNECT, OnUpdateModeShortDisconnect)
	ON_UPDATE_COMMAND_UI(ID_FILE_CLOSE, OnUpdateFileClose)
	ON_COMMAND(ID_VIEW_REFRESH, OnViewRefresh)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CGMEDoc construction/destruction

CGMEDoc::CGMEDoc()
{
	theInstance = this;

	editMode = GME_EDIT_MODE;

	modeBar.CreateEx(CMainFrame::theInstance,
        TBSTYLE_TRANSPARENT | TBSTYLE_FLAT, 
        WS_CHILD|WS_VISIBLE|CBRS_ALIGN_LEFT, 
        CRect(0,0,0,0),
        IDW_TOOLBAR_MODE); // provide unqiue ID for each toolbar [important !!!] 
                           // see MainFrm.cpp OnCreate for other details
	modeBar.LoadToolBar(IDR_TOOLBAR_MODE);
	modeBar.SetBarStyle(modeBar.GetBarStyle() 
		| CBRS_GRIPPER | CBRS_TOOLTIPS | CBRS_FLYBY | CBRS_SIZE_DYNAMIC);
	modeBar.EnableDocking(CBRS_ALIGN_ANY);
	modeBar.SetWindowText(_T("Mode")); // will show this title when floating

	CMainFrame::theInstance->ShowControlBar(&modeBar, TRUE, FALSE);
	CMainFrame::theInstance->DockControlBar(&modeBar,AFX_IDW_DOCKBAR_LEFT);

	try {
	    COMTHROW( resolver.CoCreateInstance(L"Mga.MgaResolver"));
	}
	catch(hresult_exception e) {
		AfxMessageBox("Fatal error: cannot create MGA Resolver component!",MB_OK | MB_ICONSTOP);
	}
}

CGMEDoc::~CGMEDoc()
{
	theInstance = 0;
	if (modeBar.GetSafeHwnd() && (!modeBar.IsWindowVisible())) {
		CMainFrame::theInstance->ShowControlBar(&modeBar, FALSE, FALSE);
	}
}

bool CGMEDoc::CreateFcoList(CGuiObjectList* list,CComPtr<IMgaFCOs> &fcos,CGMEView *view)
{
	try {
		view->BeginTransaction(TRANSACTION_READ_ONLY);
		ASSERT(view);

		CComPtr<IMgaFCOs> coll;
		COMTHROW(coll.CoCreateInstance(OLESTR("Mga.MgaFCOs")));
		POSITION pos = list->GetHeadPosition();
		while(pos) {
			COMTHROW(coll->Append(list->GetNext(pos)->mgaFco));
		}
		COMTHROW(view->currentModel->AddInternalConnections(coll,&fcos));
		view->CommitTransaction();
	}
	catch(hresult_exception e) {
		view->AbortTransaction(e.hr);
		return false;
	}
	return true;
}

bool CGMEDoc::CreateFcoList2(CGuiFcoList* list,CComPtr<IMgaFCOs> &fcos,CGMEView *view, long* p_howmany, bool* array[8])
{
	bool all_containers = *array[ 0 ];
	bool all_refs		= *array[ 1 ];
	bool all_sets		= *array[ 2 ];
	bool all_conns      = *array[ 3 ];

	bool any_container	= *array[ 4 ];
	bool any_ref		= *array[ 5 ];
	bool any_set		= *array[ 6 ];
	bool any_conn		= *array[ 7 ];

	long how_many = 0, now_how_many = 0;
	try {
		view->BeginTransaction(TRANSACTION_READ_ONLY);
		ASSERT(view);

		CComPtr<IMgaFCOs> coll;
		COMTHROW(coll.CoCreateInstance(OLESTR("Mga.MgaFCOs")));
		POSITION pos = list->GetHeadPosition();
		while(pos) {
			CComObjPtr<IMgaFCO> one_fco = list->GetNext(pos)->mgaFco;
			COMTHROW(coll->Append( one_fco));

			objtype_enum type_info;
			COMTHROW( one_fco->get_ObjType(&type_info)); 

			all_containers &= (type_info == OBJTYPE_MODEL || type_info == OBJTYPE_FOLDER);
			all_refs  &= (type_info == OBJTYPE_REFERENCE);
			all_sets  &= (type_info == OBJTYPE_SET);
			all_conns &= (type_info == OBJTYPE_CONNECTION);

			any_container |= (type_info == OBJTYPE_MODEL || type_info == OBJTYPE_FOLDER);
			any_ref  |= (type_info == OBJTYPE_REFERENCE);
			any_set  |= (type_info == OBJTYPE_SET);
			any_conn |= (type_info == OBJTYPE_CONNECTION);
			
			++how_many;
		}
		COMTHROW(view->currentModel->AddInternalConnections(coll,&fcos));

		// some connections might have been added
		COMTHROW( fcos->get_Count( &now_how_many) ); 
		
		if( now_how_many > how_many)
			any_conn = true;

		view->CommitTransaction();
	}
	catch(hresult_exception e) {
		view->AbortTransaction(e.hr);
		return false;
	}
	
	*p_howmany = now_how_many;

	*array[ 0 ] = all_containers;
	*array[ 1 ] = all_refs;
	*array[ 2 ] = all_sets;
	*array[ 3 ] = all_conns;
	*array[ 4 ] = any_container;
	*array[ 5 ] = any_ref;
	*array[ 6 ] = any_set;
	*array[ 7 ] = any_conn;

	return true;
}

bool CGMEDoc::CreateAnnotationList(CGuiAnnotatorList* list,CComPtr<IMgaRegNodes> &anns,CGMEView *view)
{
	try {
		view->BeginTransaction(TRANSACTION_READ_ONLY);
		ASSERT(view);
		COMTHROW(anns.CoCreateInstance(OLESTR("Mga.MgaRegNodes")));
		POSITION pos = list->GetHeadPosition();
		while(pos) {
			COMTHROW(anns->Append(list->GetNext(pos)->rootNode));
		}
		view->CommitTransaction();
	}
	catch(hresult_exception e) {
		view->AbortTransaction(e.hr);
		return false;
	}
	return true;
}

void CGMEDoc::CopyToClipboard(CGuiObjectList* list, CGuiAnnotatorList *annList,CGMEDataDescriptor* desc,CGMEView *view)
{
	ASSERT( list && desc );

	CGMEDataSource* pDataSource = new CGMEDataSource(theApp.mgaProject);
	CComPtr<IMgaFCOs> fcos;
	CreateFcoList(list,fcos,view);
	CComPtr<IMgaRegNodes> anns;
	CreateAnnotationList(annList, anns,view);
	pDataSource->SetNodes(fcos);
	pDataSource->SetAnnotations(anns);
	pDataSource->CacheDescriptor(desc);
	pDataSource->SetClipboard();
	
}

void CGMEDoc::CopyClosureToClipboard( CGuiObjectList* list, CGuiAnnotatorList *annList, CGMEDataDescriptor* desc, CGMEView *view)
{
	ASSERT( list && desc );

	CComPtr<IMgaFCOs> fcos;
	CreateFcoList( list, fcos, view);

	CComPtr<IMgaClosure> cl;

	VARIANT_BOOL _refersTo           ; VARIANT_BOOL _referredBy;
	VARIANT_BOOL _connections        ; VARIANT_BOOL _connsThroughRefport;
	VARIANT_BOOL _setMembers         ; VARIANT_BOOL _memberOfSets;
	VARIANT_BOOL _modelContainment   ; VARIANT_BOOL _partOfModels;
	VARIANT_BOOL _folderContainment  ; VARIANT_BOOL _partOfFolders;
	VARIANT_BOOL _baseTypeOf         ; VARIANT_BOOL _derivedTypesOf;
	long _libraryElementHandling;         // 0: stubs; 1: stop; 2: continue
	long _wrappingMode;                   // 0: miniproject; 1: automatic; 2: as is
	long _kindFilter;                     // -1 = all; otherwise calculate the 0b000rsamf mask; where f = Folders; m = Models; a = Atoms; s = Sets; r = References
	long _options;
	VARIANT_BOOL _directionDown;
	VARIANT_BOOL _copyToFile;             // file or clipboard
	VARIANT_BOOL _okPressed;              // how ended the dialog
	CComBSTR _userSelFileName;

	COMTHROW( cl.CoCreateInstance(L"Mga.MgaClosure") );
	COMTHROW( cl->SelectiveClosureDlg
		( VARIANT_FALSE
		, &_refersTo           , &_referredBy
		, &_connections        , &_connsThroughRefport
		, &_setMembers         , &_memberOfSets
		, &_modelContainment   , &_partOfModels
		, &_folderContainment  , &_partOfFolders
		, &_baseTypeOf         , &_derivedTypesOf
		, &_libraryElementHandling          // 0: stubs; 1: stop; 2: continue
		, &_wrappingMode                    // 0: miniproject; 1: automatic; 2: as is
		, &_kindFilter                      // -1 = all; otherwise calculate the 0b000rsamf mask; where f = Folders; m = Models; a = Atoms; s = Sets; r = References
		, &_options
		, &_directionDown
		, &_copyToFile              // file or clipboard
		, &_okPressed               // how ended the dialog
		, &_userSelFileName
		));
	if( _okPressed == VARIANT_FALSE) return;

	CComObjPtr<IMgaFCOs> sel_fcos, top_fcos;
	CComObjPtr<IMgaFolders> sel_folders, top_folders;
	VARIANT_BOOL tops_filled = VARIANT_FALSE;
	CComBstrObj acc_kinds;

	COMTHROW( cl->SelectiveClosureCalc
		( fcos          , 0
		, _refersTo          , _referredBy
		, _connections       , _connsThroughRefport
		, _setMembers        , _memberOfSets
		, _modelContainment  , _partOfModels
		, _folderContainment , _partOfFolders
		, _baseTypeOf        , _derivedTypesOf
		, _libraryElementHandling          // 0: stubs, 1: stop, 2: continue
		, _wrappingMode                    // 0: miniproject, 1: automatic, 2: as is
		, _kindFilter                      // -1 = all, otherwise calculate the 0b000rsamf mask, where f = Folders, m = Models, a = Atoms, s = Sets, r = References
		, _directionDown
		, PutOut( sel_fcos), PutOut( sel_folders)
		, PutOut( top_fcos), PutOut( top_folders)
		, &tops_filled
		, PutOut( acc_kinds)
		));

	if( !sel_fcos && !sel_folders) return;

	if( _copyToFile == VARIANT_TRUE) 
	{
		CComPtr<IMgaDumper> dumper;
		COMTHROW( dumper.CoCreateInstance(L"Mga.MgaDumper") );

		if( _wrappingMode == 0)
		{
			COMTHROW( dumper->DumpClos( sel_fcos, sel_folders, _userSelFileName, _options));
		}
		else if( _wrappingMode == 1 || _wrappingMode == 2)
		{
			if ( tops_filled == VARIANT_TRUE)
				COMTHROW( dumper->DumpClosR( sel_fcos, sel_folders, _userSelFileName, top_fcos, top_folders, _options, 0, acc_kinds));
			else 
				// dump like in miniproject case from The RootFolder
				// if the data is copied to clipboard then the DumpClosR will call the 
				// DumpClos when it will realize the topfco and topfolders are empty
				COMTHROW( dumper->DumpClos( sel_fcos, sel_folders, _userSelFileName, _options));
		}
	}
	else // _copyToFile == VARIANT_FALSE 
	{
		CGMEClosureDataSource * pDataSource = new CGMEClosureDataSource( theApp.mgaProject);

		if ( sel_fcos)		pDataSource->SetNodes( sel_fcos);
		if ( sel_folders)	pDataSource->SetFolders( sel_folders);
		
		if ( top_fcos)		pDataSource->SetTopNodes( top_fcos);
		if ( top_folders)	pDataSource->SetTopNodes( top_folders);

		pDataSource->SetOptions( _options);
		pDataSource->SetAcceptingKinds( acc_kinds);

		CComPtr<IMgaRegNodes> anns; 
		CreateAnnotationList( annList, anns, view);
		pDataSource->SetAnnotations( anns);
		pDataSource->CacheDescriptor( desc);
		pDataSource->SetClipboard();
		pDataSource->FlushClipboard();
	}
}

void CGMEDoc::CopySmartToClipboard( CGuiFcoList* list, CGuiAnnotatorList *annList, CGMEDataDescriptor* desc, CGMEView *view)
{
	int k = 0;
	bool all_containers = true;
	bool all_refs = true;
	bool all_sets = true;
	bool all_conns = true;

	bool any_container = false;
	bool any_ref = false;
	bool any_set = false;
	bool any_conn = false;

	ASSERT( list && desc );

	CComPtr<IMgaFCOs> fcos;
	bool* b_array[8] = { &all_containers, &all_refs, &all_sets, &all_conns, &any_container, &any_ref, &any_set, &any_conn };
	long number_of_elems = 0;

	if( !CreateFcoList2( list, fcos, view, &number_of_elems, b_array)) return;

	if( number_of_elems == 0) 
	{
		AfxMessageBox("No object selected", MB_ICONERROR);
		return;
	}

	
	CComPtr<IMgaClosure> cl;
	CComObjPtr<IMgaFCOs> sel_fcos, top_fcos;
	CComObjPtr<IMgaFolders> sel_folders, top_folders;
	long options;
	CComBstrObj acckinds; 
	CComBstrObj path; 

	// commented out when new smart copy introduced
	// if container selected then it must be lonely (no multiple selection)
	// if container selected no other kinds allowed
	//if( any_container && !all_containers || number_of_elems > 1 && all_containers)
	//{
	//	AfxMessageBox("Invalid selection for smart copy. You can select either one container or several non-containers.", MB_ICONERROR);
	//	return;
	//}

	if( any_container) k |= 1024;

	if( any_ref) k |= 2048;
	if( any_set) k |= 4096;
	if( any_conn) k |= 8192;

	COMTHROW( cl.CoCreateInstance(L"Mga.MgaClosure") );
	
	COMTHROW( cl->SmartCopy( 
		fcos, 
		0, 
		PutOut( sel_fcos), 
		PutOut( sel_folders), 
		PutOut( top_fcos), 
		PutOut( top_folders), 
		k, 
		&options, 
		PutOut( acckinds),
		PutOut( path)
		));
	
	if ( sel_fcos || sel_folders) // if both null the operation is not needed (either dumped to file already or unsuccessful)
	{
		CRectList dummyAnnList;

		CGMEClosureDataSource * pDataSource = new CGMEClosureDataSource( theApp.mgaProject);

		if ( sel_fcos)		pDataSource->SetNodes( sel_fcos);
		if ( sel_folders)	pDataSource->SetFolders( sel_folders);
		
		if ( top_fcos)		pDataSource->SetTopNodes( top_fcos);
		if ( top_folders)	pDataSource->SetTopNodes( top_folders);

		pDataSource->SetOptions( options);
		pDataSource->SetAbsPathPart( path);
		pDataSource->SetAcceptingKinds( acckinds);

		CComPtr<IMgaRegNodes> anns; 
		CreateAnnotationList( annList, anns, view);
		pDataSource->SetAnnotations( anns);
		pDataSource->CacheDescriptor( desc);
		pDataSource->SetClipboard();
		pDataSource->FlushClipboard();
	}
}

DROPEFFECT CGMEDoc::DoDragDrop(CGuiObjectList* list, CGuiAnnotatorList *annList, CGMEDataDescriptor* desc,
	DWORD dwEffects, LPCRECT lpRectStartDrag,CGMEView *view)
{
	ASSERT( list && desc && annList );

	CGMEDataSource dataSource(theApp.mgaProject);
	CComPtr<IMgaFCOs> fcos;
	CreateFcoList(list,fcos,view);
	CComPtr<IMgaRegNodes> anns;
	CreateAnnotationList(annList, anns,view);
	dataSource.SetNodes(fcos);
	dataSource.SetAnnotations(anns);
	dataSource.CacheDescriptor(desc);
	DROPEFFECT de = dataSource.DoDragDrop(dwEffects,lpRectStartDrag);
	return de;
}

DROPEFFECT CGMEDoc::DoDragDrop(CGuiObject *guiObj, CGMEDataDescriptor* desc,
	DWORD dwEffects, LPCRECT lpRectStartDrag)
{
	CGMEDataSource dataSource(theApp.mgaProject);
	dataSource.SetMetaRole(guiObj->metaRole);
	dataSource.CacheDescriptor(desc);
	DROPEFFECT de = dataSource.DoDragDrop(dwEffects,lpRectStartDrag);
	return de;
}

CGMEView *CGMEDoc::FindView(CComPtr<IMgaModel> model)
{
	if(model != 0) {
		POSITION pos = GetFirstViewPosition();
		while (pos != NULL) {
			CGMEView* view = (CGMEView *)GetNextView(pos);
			VARIANT_BOOL b; 
			// COMTHROW(model->get_IsEqual(view->GetCurrentModel(), &b));  Territory problems
			// bool b = model.IsEqualObject(view->GetCurrentModel());
			view->BeginTransaction(TRANSACTION_READ_ONLY);
			COMTHROW(view->GetCurrentModel()->get_IsEqual(model, &b));
			view->CommitTransaction();
			if(b) 
				return view;
		}
	}
	return 0;
}

void CGMEDoc::InvalidateAllViews(bool thorough,bool /*fullAutoRoute*/)
{
	POSITION pos = GetFirstViewPosition();
	while (pos != NULL) {
		CGMEView* pView = (CGMEView *)GetNextView(pos);
/*
		if(!pView->IsIconic())
			fullAutoRoute ? pView->AutoRoute() : pView->IncrementalAutoRoute();
*/
		pView->Invalidate(thorough);
	}
	if(thorough) {
		CMDIFrameWnd *pFrame = (CMDIFrameWnd*)AfxGetApp()->m_pMainWnd;
		CMDIChildWnd *pChild = (CMDIChildWnd *) pFrame->GetActiveFrame();
		((CGMEView *)(pChild->GetActiveView()))->Invalidate(thorough);
	}
}

void CGMEDoc::ShowObject(CComPtr<IUnknown> alienObject, BOOL inParent)
{
	// Be prepared for in-transaction and out-transaction scenarios
	// mgaFCO must be transfered to current territory
		
	bool isModel = true;
	if(alienObject == NULL) {
		return;
	}

	if (!theApp.mgaProject) {
		return;
	}

	CComPtr<IMgaFCO> fco;
	CComPtr<IMgaModel> model;
	
	
	long status;
	COMTHROW(theApp.mgaProject->get_ProjectStatus(&status));
	bool inTrans = (status & 0x08L) != 0;
	CComPtr<IMgaTerritory> terr;

	if (!SUCCEEDED(alienObject.QueryInterface(&model))) {  // Alien object is not a modell
		isModel = false;
		if (SUCCEEDED(alienObject.QueryInterface(&fco))) {
			
			try {
				if (!inTrans) {
					COMTHROW(theApp.mgaProject->CreateTerritory(NULL, &terr));
					COMTHROW(theApp.mgaProject->BeginTransaction(terr));
				}
				else {
					COMTHROW(theApp.mgaProject->get_ActiveTerritory(&terr));
				}
				CComPtr<IMgaFCO> tmpfco = fco;
				fco = NULL;
				COMTHROW(terr->OpenFCO(tmpfco, &fco));
			}
			catch (hresult_exception e) {
				return;
			}

			try {
				CComPtr<IMgaObject> parent;
				COMTHROW(fco->GetParent(&parent));
				model = NULL;
				COMTHROW(parent.QueryInterface(&model));
			}
			catch (hresult_exception e) {
				model = NULL;
				if (!inTrans) 
					theApp.mgaProject->AbortTransaction();
			}

			if (!inTrans) {
				theApp.mgaProject->CommitTransaction();
			}
		}
	}

	if (model  &&  isModel  &&  inParent)
	{

		try 
		{
			if (!inTrans) {
				COMTHROW(theApp.mgaProject->CreateTerritory(NULL, &terr));
				COMTHROW(theApp.mgaProject->BeginTransaction(terr));
				CComPtr<IMgaModel> tmpmod = model; // put model into the active territory
				model = NULL;
				CComPtr<IMgaFCO> fco_mod;
				COMTHROW(terr->OpenFCO(tmpmod, &fco_mod)); // OpenFCO fills an IMgaFCO
				COMTHROW( fco_mod.QueryInterface( &model));// casted to IMgaModel
			}
			CComPtr<IMgaObject> parent;
			COMTHROW(model->GetParent(&parent));
			COMTHROW(model.QueryInterface(&fco));
			model = NULL;
			COMTHROW(parent.QueryInterface(&model));
			// it might be invoked from the ActiveBrowser (on a top model)
			// in which case line above must throw if parent is a folder

			if (!inTrans) 
				theApp.mgaProject->CommitTransaction();
		}
		catch (hresult_exception e) 
		{
			model = NULL;
			if (!inTrans) 
				theApp.mgaProject->AbortTransaction();
		}
	}
	if (model) {
		CGMEView *view  = NULL;

		if (!theApp.multipleView) {
			view = FindView(model);
			if(!view) {
				SetNextToView(model,"", fco);
			}
			else {
				view->SetCenterObject(fco);
			}
		}
		else {
			SetNextToView(model,"", fco);
		}

		CMainFrame::theInstance->CreateNewView(view);
		// PETER: This is needed to get the focus (SetFocus does not work, since it uses SendMessage())
		::PostMessage(CMainFrame::theInstance->GetSafeHwnd(), WM_SETFOCUS, 0, 0);
	}
}

void CGMEDoc::ShowObject(LPCTSTR objectID)
{
	CComBSTR bstrID(objectID);
	CComPtr<IMgaObject> object;
	CComPtr<IUnknown> unk;
	bool inTrans = false;


	try {
		long status;
		COMTHROW(theApp.mgaProject->get_ProjectStatus(&status));
		inTrans = (status & 0x08L) != 0;
		CComPtr<IMgaTerritory> terr;
		if (!inTrans) {
			COMTHROW(theApp.mgaProject->CreateTerritory(NULL, &terr));
			COMTHROW(theApp.mgaProject->BeginTransaction(terr));
		}
		else {
			COMTHROW(theApp.mgaProject->get_ActiveTerritory(&terr));
		}

		COMTHROW(theApp.mgaProject->GetObjectByID(bstrID, &object));
		COMTHROW(object.QueryInterface(&unk));

		if (!inTrans) {
			theApp.mgaProject->CommitTransaction();
		}
	}
	catch (hresult_exception e) {
		if (!inTrans) 
			theApp.mgaProject->AbortTransaction();
		return;
	}

	
	ShowObject(unk);
}

int  CGMEDoc::HowManyViewsAreOpen()
{
	int i = 0;
	POSITION pos = GetFirstViewPosition();
	if(pos) {
		while (pos != NULL) {
			CGMEView* pView = (CGMEView *)GetNextView(pos);
			++i;
		}
	}
	return i;
}

void CGMEDoc::ResetAllViews()
{
	POSITION pos = GetFirstViewPosition();
	if(pos) {
		while (pos != NULL) {
			CGMEView* pView = (CGMEView *)GetNextView(pos);
			pView->Reset(true);
		}
//		InvalidateAllViews(true);
	}
}

void CGMEDoc::ChangeAspects(CString aspName)
{
	POSITION pos = GetFirstViewPosition();
	if(pos) {
		while (pos != NULL) {
			CGMEView* pView = (CGMEView *)GetNextView(pos);
			pView->ChangeAspect(aspName);
		}
	}
}

void CGMEDoc::ViewModeChange()
{
	POSITION pos = GetFirstViewPosition();
	while (pos != NULL) {
		CGMEView* pView = (CGMEView *)GetNextView(pos);
		pView->ModeChange();
	}

	// make sure that the currently open model has the focus
	CMainFrame::theInstance->SetFocus();
}

/////////////////////////////////////////////////////////////////////////////
// CGMEDoc diagnostics

#ifdef _DEBUG
void CGMEDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CGMEDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CGMEDoc commands

void CGMEDoc::SetMode(long mode)
{
	switch (mode)
	{
	case 0: // edit
		OnModeEdit();
		break;
	case 1: // connect
		OnModeAutoconnect();
		break;
	case 2: // disconnect
		OnModeDisconnect();
		break;
	case 3: // set
		OnModeSet();
		break;
	case 4: // zoom
		OnModeZoom();
		break;
	case 5: // visualize
		OnModeVisualize();
		break;
	case 6: // short-connect
		OnModeShortConnect();
		break;
	case 7: // short-disconnect
		OnModeShortDisconnect();
		break;
	}
}

void CGMEDoc::OnModeAutoconnect() 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnModeAutoconnect\r\n");
	editMode = GME_AUTOCONNECT_MODE;
	CMainFrame::theInstance->WriteStatusMode("AUTOCONNECT");
	ViewModeChange();
}

void CGMEDoc::OnUpdateModeAutoconnect(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(editMode == GME_AUTOCONNECT_MODE);
}

void CGMEDoc::OnModeDisconnect() 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnModeDisconnect\r\n");
	editMode = GME_DISCONNECT_MODE;
	CMainFrame::theInstance->WriteStatusMode("DISCONNECT");
	ViewModeChange();
}

void CGMEDoc::OnUpdateModeDisconnect(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(editMode == GME_DISCONNECT_MODE);
}

void CGMEDoc::OnModeEdit() 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnModeEdit\r\n");
	editMode = GME_EDIT_MODE;
	CMainFrame::theInstance->WriteStatusMode("EDIT");
	ViewModeChange();
}

void CGMEDoc::OnUpdateModeEdit(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(editMode == GME_EDIT_MODE);
}

void CGMEDoc::OnModeSet() 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnModeSet\r\n");
	editMode = GME_SET_MODE;
	CMainFrame::theInstance->WriteStatusMode("SET");
	ViewModeChange();
}

void CGMEDoc::OnUpdateModeSet(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(editMode == GME_SET_MODE);
}

void CGMEDoc::OnModeZoom() 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnModeZoom\r\n");
	editMode = GME_ZOOM_MODE;
	CMainFrame::theInstance->WriteStatusMode("ZOOM");
	ViewModeChange();
}

void CGMEDoc::OnUpdateModeZoom(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(editMode == GME_ZOOM_MODE);
}

void CGMEDoc::OnModeVisualize() 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnModeVisualize\r\n");
	editMode = GME_VISUAL_MODE;
	CMainFrame::theInstance->WriteStatusMode("VISUALIZE");
	ViewModeChange();
}

void CGMEDoc::OnUpdateModeVisualize(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(editMode == GME_VISUAL_MODE);
}

void CGMEDoc::OnModeShortConnect()
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnModeShortConnect\r\n");
	editMode = GME_SHORTAUTOCONNECT_MODE;
	CMainFrame::theInstance->WriteStatusMode("AUTOCONNECT");
	ViewModeChange();
}

void CGMEDoc::OnUpdateModeShortConnect(CCmdUI* pCmdUI)
{
	pCmdUI->SetCheck(editMode == GME_SHORTAUTOCONNECT_MODE);
}

void CGMEDoc::OnModeShortDisconnect()
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnModeShortDisconnect\r\n");
	editMode = GME_SHORTDISCONNECT_MODE;
	CMainFrame::theInstance->WriteStatusMode("DISCONNECT");
	ViewModeChange();
}

void CGMEDoc::OnUpdateModeShortDisconnect(CCmdUI* pCmdUI)
{
	pCmdUI->SetCheck(editMode == GME_SHORTDISCONNECT_MODE);
}

void CGMEDoc::OnUpdateFileClose(CCmdUI* pCmdUI) 
{
	POSITION pos = GetFirstViewPosition();
	pCmdUI->Enable(pos != NULL);
}

void CGMEDoc::OnCloseDocument(bool suppressErrors) 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnCloseDocument\r\n");
	POSITION pos = GetFirstViewPosition();
	while (pos) {
		CGMEView* view = dynamic_cast<CGMEView*>(GetNextView(pos));
		ASSERT(view);
		bool canClose = view->SendCloseModelEvent();
		if ((!suppressErrors) && (!canClose)) {
			// View cannot be closed, eg.: constraint violations
			return;
		}
	}

	CDocument::OnCloseDocument();
	resolver = NULL;
	theApp.CloseProject();
	delete this;
}

BOOL CGMEDoc::OnNewDocument()
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnNewDocument\r\n");
	if (!CDocument::OnNewDocument())
		return FALSE;

	// TODO: add reinitialization code here
	// (SDI documents will reuse this document)

	return TRUE;
}

BOOL CGMEDoc::OnOpenDocument(LPCTSTR /*lpszPathName*/) 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnOpenDocument\r\n");
	if (!CDocument::OnNewDocument())
		return FALSE;
	
	// TODO: Add your specialized creation code here
	
	return TRUE;
}

void CGMEDoc::OnViewRefresh() 
{
	CGMEEventLogger::LogGMEEvent("CGMEDoc::OnViewRefresh\r\n");
	ResetAllViews();
	CGMEBrowser::theInstance->RefreshAll();
}
